﻿Option Strict On

Imports System.Text
Imports System.Net.Sockets
Imports System.Threading
Imports System.Globalization
Imports System.ComponentModel

Namespace Proxy
    ''' <summary>
    ''' HTTP connection proxy class.  This class implements the HTTP standard proxy protocol.
    ''' <para>
    ''' You can use this class to set up a connection to an HTTP proxy server.  Calling the 
    ''' CreateConnection() method initiates the proxy connection and return a standard
    ''' System.Net.Socks.TcpClient object that can be used as normal.  The proxy plumbing
    ''' is all handled for you.
    ''' </para>
    ''' <code>
    ''' 
    ''' </code>
    ''' </summary>
    Public Class HttpProxyClient
        Implements IProxyClient
        Private _proxyHost As String
        Private _proxyPort As Integer
        Private _respCode As HttpResponseCodes
        Private _respText As String
        Private _tcpClient As TcpClient

        Private Const HTTP_PROXY_DEFAULT_PORT As Integer = 8080
        Private Const HTTP_PROXY_CONNECT_CMD As String = "CONNECT {0}:{1} HTTP/1.0 " & vbCr & vbLf & "HOST {0}:{1}" & vbCr & vbLf & vbCr & vbLf
        Private Const WAIT_FOR_DATA_INTERVAL As Integer = 50
        ' 50 ms
        Private Const WAIT_FOR_DATA_TIMEOUT As Integer = 15000
        ' 15 seconds
        Private Const PROXY_NAME As String = "HTTP"

        Private Enum HttpResponseCodes
            None = 0
            [Continue] = 100
            SwitchingProtocols = 101
            OK = 200
            Created = 201
            Accepted = 202
            NonAuthoritiveInformation = 203
            NoContent = 204
            ResetContent = 205
            PartialContent = 206
            MultipleChoices = 300
            MovedPermanetly = 301
            Found = 302
            SeeOther = 303
            NotModified = 304
            UserProxy = 305
            TemporaryRedirect = 307
            BadRequest = 400
            Unauthorized = 401
            PaymentRequired = 402
            Forbidden = 403
            NotFound = 404
            MethodNotAllowed = 405
            NotAcceptable = 406
            ProxyAuthenticantionRequired = 407
            RequestTimeout = 408
            Conflict = 409
            Gone = 410
            PreconditionFailed = 411
            RequestEntityTooLarge = 413
            RequestURITooLong = 414
            UnsupportedMediaType = 415
            RequestedRangeNotSatisfied = 416
            ExpectationFailed = 417
            InternalServerError = 500
            NotImplemented = 501
            BadGateway = 502
            ServiceUnavailable = 503
            GatewayTimeout = 504
            HTTPVersionNotSupported = 505
        End Enum

        ''' <summary>
        ''' Constructor.
        ''' </summary>
        Public Sub New()
        End Sub

        ''' <summary>
        ''' Creates a HTTP proxy client object using the supplied TcpClient object connection.
        ''' </summary>
        ''' <param name="tcpClient">A TcpClient connection object.</param>
        Public Sub New(ByVal tcpClient As TcpClient)
            If tcpClient Is Nothing Then
                Throw New ArgumentNullException("tcpClient")
            End If

            _tcpClient = tcpClient
        End Sub


        ''' <summary>
        ''' Constructor.  The default HTTP proxy port 8080 is used.
        ''' </summary>
        ''' <param name="proxyHost">Host name or IP address of the proxy.</param>
        Public Sub New(ByVal proxyHost As String)
            If String.IsNullOrEmpty(proxyHost) Then
                Throw New ArgumentNullException("proxyHost")
            End If

            _proxyHost = proxyHost
            _proxyPort = HTTP_PROXY_DEFAULT_PORT
        End Sub

        ''' <summary>
        ''' Constructor.  
        ''' </summary>
        ''' <param name="proxyHost">Host name or IP address of the proxy server.</param>
        ''' <param name="proxyPort">Port number for the proxy server.</param>
        Public Sub New(ByVal proxyHost As String, ByVal proxyPort As Integer)
            If String.IsNullOrEmpty(proxyHost) Then
                Throw New ArgumentNullException("proxyHost")
            End If

            If proxyPort <= 0 OrElse proxyPort > 65535 Then
                Throw New ArgumentOutOfRangeException("proxyPort", "port must be greater than zero and less than 65535")
            End If

            _proxyHost = proxyHost
            _proxyPort = proxyPort
        End Sub

        ''' <summary>
        ''' Gets or sets host name or IP address of the proxy server.
        ''' </summary>
        Public Property ProxyHost() As String Implements SFDL.Proxy.Proxy.IProxyClient.ProxyHost
            Get
                Return _proxyHost
            End Get
            Set(ByVal value As String)
                _proxyHost = value
            End Set
        End Property

        ''' <summary>
        ''' Gets or sets port number for the proxy server.
        ''' </summary>
        Public Property ProxyPort() As Integer Implements SFDL.Proxy.Proxy.IProxyClient.ProxyPort
            Get
                Return _proxyPort
            End Get
            Set(ByVal value As Integer)
                _proxyPort = value
            End Set
        End Property

        ''' <summary>
        ''' Gets String representing the name of the proxy. 
        ''' </summary>
        ''' <remarks>This property will always return the value 'HTTP'</remarks>
        Public ReadOnly Property ProxyName() As String Implements SFDL.Proxy.Proxy.IProxyClient.ProxyName
            Get
                Return PROXY_NAME
            End Get
        End Property

        ''' <summary>
        ''' Gets or sets the TcpClient object. 
        ''' This property can be set prior to executing CreateConnection to use an existing TcpClient connection.
        ''' </summary>
        Public Property TcpClient() As TcpClient Implements SFDL.Proxy.Proxy.IProxyClient.TcpClient
            Get
                Return _tcpClient
            End Get
            Set(ByVal value As TcpClient)
                _tcpClient = value
            End Set
        End Property


        ''' <summary>
        ''' Creates a remote TCP connection through a proxy server to the destination host on the destination port.
        ''' </summary>
        ''' <param name="destinationHost">Destination host name or IP address.</param>
        ''' <param name="destinationPort">Port number to connect to on the destination host.</param>
        ''' <return>
        ''' return an open TcpClient object that can be used normally to communicate
        ''' with the destination server
        ''' </return>
        ''' <remarks>
        ''' This method creates a connection to the proxy server and instructs the proxy server
        ''' to make a pass through connection to the specified destination host on the specified
        ''' port.  
        ''' </remarks>
        Public Function CreateConnection(ByVal destinationHost As String, ByVal destinationPort As Integer) As TcpClient Implements SFDL.Proxy.Proxy.IProxyClient.CreateConnection
            Try
                ' if we have no connection, create one
                If _tcpClient Is Nothing Then
                    If String.IsNullOrEmpty(_proxyHost) Then
                        Throw New ProxyException("ProxyHost property must contain a value.")
                    End If

                    If _proxyPort <= 0 OrElse _proxyPort > 65535 Then
                        Throw New ProxyException("ProxyPort value must be greater than zero and less than 65535")
                    End If

                    '  create new tcp client object to the proxy server
                    _tcpClient = New TcpClient()

                    ' attempt to open the connection
                    _tcpClient.Connect(_proxyHost, _proxyPort)
                End If

                '  send connection command to proxy host for the specified destination host and port
                SendConnectionCommand(destinationHost, destinationPort)

                ' return the open proxied tcp client object to the caller for normal use
                Return _tcpClient
            Catch ex As SocketException
                Throw New ProxyException(String.Format(CultureInfo.InvariantCulture, "Connection to proxy host {0} on port {1} failed.", Utils.GetHost(_tcpClient), Utils.GetPort(_tcpClient)), ex)
            End Try
        End Function


        Private Sub SendConnectionCommand(ByVal host As String, ByVal port As Integer)
            Dim stream As NetworkStream = _tcpClient.GetStream()

            ' PROXY SERVER REQUEST
            ' =======================================================================
            'CONNECT starksoft.com:443 HTTP/1.0 <CR><LF>
            'HOST starksoft.com:443<CR><LF>
            '[... other HTTP header lines ending with <CR><LF> if required]>
            '<CR><LF>    // Last Empty Line

            Dim connectCmd As String = String.Format(CultureInfo.InvariantCulture, HTTP_PROXY_CONNECT_CMD, host, port.ToString(CultureInfo.InvariantCulture))
            Dim request As Byte() = ASCIIEncoding.ASCII.GetBytes(connectCmd)

            ' send the connect request
            stream.Write(request, 0, request.Length)

            ' wait for the proxy server to respond
            WaitForData(stream)

            ' PROXY SERVER RESPONSE
            ' =======================================================================
            'HTTP/1.0 200 Connection Established<CR><LF>
            '[.... other HTTP header lines ending with <CR><LF>..
            'ignore all of them]
            '<CR><LF>    // Last Empty Line

            ' create an byte response array  
            Dim response As Byte() = New Byte(_tcpClient.ReceiveBufferSize - 1) {}
            Dim sbuilder As New StringBuilder()
            Dim bytes As Integer = 0
            Dim total As Long = 0

            Do
                bytes = stream.Read(response, 0, _tcpClient.ReceiveBufferSize)
                total += bytes
                sbuilder.Append(System.Text.ASCIIEncoding.UTF8.GetString(response, 0, bytes))
            Loop While stream.DataAvailable

            ParseResponse(sbuilder.ToString())

            '  evaluate the reply code for an error condition
            If _respCode <> HttpResponseCodes.OK Then
                HandleProxyCommandError(host, port)
            End If
        End Sub

        Private Sub HandleProxyCommandError(ByVal host As String, ByVal port As Integer)
            Dim msg As String

            Select Case _respCode
                Case HttpResponseCodes.None
                    msg = String.Format(CultureInfo.InvariantCulture, "Proxy destination {0} on port {1} failed to return a recognized HTTP response code.  Server response: {2}", Utils.GetHost(_tcpClient), Utils.GetPort(_tcpClient), _respText)
                    Exit Select

                Case HttpResponseCodes.BadGateway
                    'HTTP/1.1 502 Proxy Error (The specified Secure Sockets Layer (SSL) port is not allowed. ISA Server is not configured to allow SSL requests from this port. Most Web browsers use port 443 for SSL requests.)
                    msg = String.Format(CultureInfo.InvariantCulture, "Proxy destination {0} on port {1} responded with a 502 code - Bad Gateway.  If you are connecting to a Microsoft ISA destination please refer to knowledge based article Q283284 for more information.  Server response: {2}", Utils.GetHost(_tcpClient), Utils.GetPort(_tcpClient), _respText)
                    Exit Select
                Case Else

                    msg = String.Format(CultureInfo.InvariantCulture, "Proxy destination {0} on port {1} responded with a {2} code - {3}", Utils.GetHost(_tcpClient), Utils.GetPort(_tcpClient), CInt(_respCode).ToString(CultureInfo.InvariantCulture), _respText)
                    Exit Select
            End Select

            '  throw a new application exception 
            Throw New ProxyException(msg)
        End Sub

        Private Sub WaitForData(ByVal stream As NetworkStream)
            Dim sleepTime As Integer = 0
            While Not stream.DataAvailable
                Thread.Sleep(WAIT_FOR_DATA_INTERVAL)
                sleepTime += WAIT_FOR_DATA_INTERVAL
                If sleepTime > WAIT_FOR_DATA_TIMEOUT Then
                    Throw New ProxyException(String.Format("A timeout while waiting for the proxy server at {0} on port {1} to respond.", Utils.GetHost(_tcpClient), Utils.GetPort(_tcpClient)))
                End If
            End While
        End Sub

        Private Sub ParseResponse(ByVal response As String)
            Dim data As String() = Nothing

            '  get rid of the LF character if it exists and then split the string on all CR
            data = response.Replace(ControlChars.Lf, " "c).Split(ControlChars.Cr)

            ParseCodeAndText(data(0))
        End Sub

        Private Sub ParseCodeAndText(ByVal line As String)
            Dim begin As Integer = 0
            Dim [end] As Integer = 0
            Dim val As String = Nothing

            If line.IndexOf("HTTP") = -1 Then
                Throw New ProxyException(String.Format("No HTTP response received from proxy destination.  Server response: {0}.", line))
            End If

            begin = line.IndexOf(" ") + 1
            [end] = line.IndexOf(" ", begin)

            val = line.Substring(begin, [end] - begin)
            Dim code As Int32 = 0

            If Not Int32.TryParse(val, code) Then
                Throw New ProxyException(String.Format("An invalid response code was received from proxy destination.  Server response: {0}.", line))
            End If

            _respCode = CType(code, HttpResponseCodes)
            _respText = line.Substring([end] + 1).Trim()
        End Sub



#Region "Async Methods"

        Private _asyncWorker As BackgroundWorker
        Private _asyncException As Exception
        Private _asyncCancelled As Boolean

        ''' <summary>
        ''' Gets a value indicating whether an asynchronous operation is running.
        ''' </summary>
        ''' <remarks>return true if an asynchronous operation is running; otherwise, false.
        ''' </remarks>
        Public ReadOnly Property IsBusy() As Boolean
            Get
                Return If(_asyncWorker Is Nothing, False, _asyncWorker.IsBusy)
            End Get
        End Property

        ''' <summary>
        ''' Gets a value indicating whether an asynchronous operation is cancelled.
        ''' </summary>
        ''' <remarks>return true if an asynchronous operation is cancelled; otherwise, false.
        ''' </remarks>
        Public ReadOnly Property IsAsyncCancelled() As Boolean
            Get
                Return _asyncCancelled
            End Get
        End Property

        ''' <summary>
        ''' Cancels any asychronous operation that is currently active.
        ''' </summary>
        Public Sub CancelAsync()
            If _asyncWorker IsNot Nothing AndAlso Not _asyncWorker.CancellationPending AndAlso _asyncWorker.IsBusy Then
                _asyncCancelled = True
                _asyncWorker.CancelAsync()
            End If
        End Sub

        Private Sub CreateAsyncWorker()
            If _asyncWorker IsNot Nothing Then
                _asyncWorker.Dispose()
            End If
            _asyncException = Nothing
            _asyncWorker = Nothing
            _asyncCancelled = False
            _asyncWorker = New BackgroundWorker()
        End Sub

        ''' <summary>
        ''' Event handler for CreateConnectionAsync method completed.
        ''' </summary>
        Public Event CreateConnectionAsyncCompleted As EventHandler(Of CreateConnectionAsyncCompletedEventArgs) Implements SFDL.Proxy.Proxy.IProxyClient.CreateConnectionAsyncCompleted

        ''' <summary>
        ''' Asynchronously creates a remote TCP connection through a proxy server to the destination host on the destination port.
        ''' </summary>
        ''' <param name="destinationHost">Destination host name or IP address.</param>
        ''' <param name="destinationPort">Port number to connect to on the destination host.</param>
        ''' <remarks>
        ''' This method creates a connection to the proxy server and instructs the proxy server
        ''' to make a pass through connection to the specified destination host on the specified
        ''' port.  
        ''' </remarks>
        Public Sub CreateConnectionAsync(ByVal destinationHost As String, ByVal destinationPort As Integer) Implements SFDL.Proxy.Proxy.IProxyClient.CreateConnectionAsync
            If _asyncWorker IsNot Nothing AndAlso _asyncWorker.IsBusy Then
                Throw New InvalidOperationException("The HttpProxy object is already busy executing another asynchronous operation.  You can only execute one asychronous method at a time.")
            End If

            CreateAsyncWorker()
            _asyncWorker.WorkerSupportsCancellation = True
            AddHandler _asyncWorker.DoWork, New DoWorkEventHandler(AddressOf CreateConnectionAsync_DoWork)
            AddHandler _asyncWorker.RunWorkerCompleted, New RunWorkerCompletedEventHandler(AddressOf CreateConnectionAsync_RunWorkerCompleted)
            Dim args As [Object]() = New [Object](1) {}
            args(0) = destinationHost
            args(1) = destinationPort
            _asyncWorker.RunWorkerAsync(args)
        End Sub

        Private Sub CreateConnectionAsync_DoWork(ByVal sender As Object, ByVal e As DoWorkEventArgs)
            Try
                Dim args As [Object]() = DirectCast(e.Argument, [Object]())
                e.Result = CreateConnection(DirectCast(args(0), String), CInt(args(1)))
            Catch ex As Exception
                _asyncException = ex
            End Try
        End Sub

        Private Sub CreateConnectionAsync_RunWorkerCompleted(ByVal sender As Object, ByVal e As RunWorkerCompletedEventArgs)
            RaiseEvent CreateConnectionAsyncCompleted(Me, New CreateConnectionAsyncCompletedEventArgs(_asyncException, _asyncCancelled, DirectCast(e.Result, TcpClient)))
        End Sub



#End Region

    End Class
End Namespace